"""Test cases asserting the results consistency between the mono and ckd modes."""
import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.scenes.spectra import MultiDeltaSpectrum
from eradiate.test_tools import util as tu


def init_mono_experiment(wavelengths, spp, reflectance, zeniths):
    """
    Initialize the monochromatic experiment.
    """
    eradiate.set_mode("mono_double")

    return eradiate.experiments.AtmosphereExperiment(
        measures=[
            eradiate.scenes.measure.MultiDistantMeasure.from_viewing_angles(
                srf=MultiDeltaSpectrum(wavelengths=wavelengths),
                zeniths=zeniths,
                azimuths=0.0 * ureg.deg,
                spp=spp,
            )
        ],
        illumination=eradiate.scenes.illumination.DirectionalIllumination(
            zenith=50.0 * ureg.deg, azimuth=0.0 * ureg.deg
        ),
        atmosphere=eradiate.scenes.atmosphere.MolecularAtmosphere.afgl_1986(
            id="molecules",
            has_absorption=True,
            has_scattering=True,
            absorption_data_sets={},
        ),
        surface=eradiate.scenes.bsdfs.LambertianBSDF(reflectance=reflectance),
    )


def init_ckd_experiment(bin_set, bins, spp, reflectance, zeniths):
    """
    Initialise CKD experiment.
    """
    eradiate.set_mode("ckd_double")

    return eradiate.experiments.AtmosphereExperiment(
        measures=[
            eradiate.scenes.measure.MultiDistantMeasure.from_viewing_angles(
                srf=MultiDeltaSpectrum(
                    wavelengths=np.array(bins, dtype=float) * ureg.nm
                ),
                zeniths=zeniths,
                azimuths=0.0 * ureg.deg,
                spp=spp,
            )
        ],
        illumination=eradiate.scenes.illumination.DirectionalIllumination(
            zenith=50.0 * ureg.deg, azimuth=0.0 * ureg.deg
        ),
        atmosphere=eradiate.scenes.atmosphere.MolecularAtmosphere.afgl_1986(
            id="molecules",
            has_absorption=True,
            has_scattering=True,
        ),
        surface=eradiate.scenes.surface.LambertianSurface(
            reflectance=reflectance,
        ),
    )


def make_figure(
    ckd_results, mono_results_averaged, fname_plot, wavelength_bin, reflectance
):
    fig = plt.figure(figsize=(8, 5))
    brf_mono_averaged = mono_results_averaged.brf.squeeze()
    brf_ckd = ckd_results.brf.squeeze()
    brf_mono_averaged.plot(x="vza", ls="dotted", marker=".")
    brf_ckd.plot(x="vza", ls="dotted", marker=".")

    # Compute relative absolute difference and annotate plot with min, max and
    # mean values
    # NB: RAD(x, y) = abs(x - y) / x
    rad = np.divide(
        np.abs(brf_mono_averaged.values - brf_ckd.values),
        brf_mono_averaged.values,
        where=brf_mono_averaged.values != 0.0,
        out=np.full_like(brf_mono_averaged.values, fill_value=np.nan),
    )
    rmad = np.mean(rad[~np.isnan(rad)])
    minrad = np.min(rad[~np.isnan(rad)])
    maxrad = np.max(rad[~np.isnan(rad)])
    text = (
        f"mean(RAD) >= {round(100 * rmad, 2)} %\n"
        f"min(RAD) >= {round(100 * minrad, 2)} %\n"
        f"max(RAD) >= {round(100 * maxrad, 2)} %"
    )
    plt.text(
        0.5,
        0.9,
        text,
        ha="center",
        va="center",
        transform=plt.gca().transAxes,
    )

    plt.legend(["mono_double (averaged)", "ckd_double"])
    plt.title(f"Wavelength bin: {wavelength_bin}, reflectance = {reflectance}")
    plt.tight_layout()
    fig.savefig(fname_plot, dpi=200)
    plt.close()


@pytest.mark.parametrize("reflectance", [0.0, 0.5, 1.0])
@pytest.mark.slow
def test_550(reflectance, artefact_dir):
    r"""
    Result consistency between ``mono_double`` and ``ckd_double`` modes
    ===================================================================

    This test checks the results consistency between mono_double and ckd_double
    modes in the [549.5, 550.5] nm wavelength bin.

    Rationale
    ---------

    * Sensor: Distant measure covering a plane (11 angular points,
      10000 sample per pixel) and targeting (0, 0, 0).
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta = 50.0°`.
    * Surface: a square surface with a Lambertian BRDF with reflectance
      parameter in :math:`\rho \in [0.0, 0.5, 1.0]`.
    * Atmosphere: a molecular atmosphere derived from the AFGL (1986)
      thermophysical properties ``us_standard`` model, with both scattering and
      absorption enabled.
    * Integrator: volumetric path tracer.

    Expected behaviour
    ------------------

    The [545, 555] nm - integrated BRF results obtained with the ``mono_double``
    and ``ckd_double`` modes should agree within 5 %.

    The ``mono_double`` BRF results are spectrally averaged according to the
    following formula:

    .. math::

       \hat{q} = \frac{
          \int_{
             \lambda_{\mathrm{min}}
          }^{
             \lambda_{\mathrm{max}}
          }
          q(\lambda) \, \mathrm{d} \lambda
       }{
          \int_{
             \lambda_{\mathrm{min}}
          }^{
             \lambda_{\mathrm{max}}
          }
          \mathrm{d} \lambda
       }

    where

    * :math:`q` is some spectral quantity (e.g. the BRF)
    * :math:`\lambda` denotes the wavelength
    * :math:`[\lambda_{\mathrm{min}},\, \lambda_{\mathrm{max}}]` is the
      wavelength bin

    Results
    -------

    Note: :math:`\mathrm{RAD}(x, y) = \left| \dfrac{x - y}{x} \right|`

    .. image:: generated/plots/test_onedim_ckd_vs_mono_550_0.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_ckd_vs_mono_550_0.5.png
       :width: 95%
    .. image:: generated/plots/test_onedim_ckd_vs_mono_550_1.0.png
       :width: 95%
    """
    # Settings
    zeniths = np.linspace(-75, 75, 11)
    spp = 1e4
    wavelength_min = 549.5 * ureg.nm
    wavelength_max = 550.5 * ureg.nm
    wavenumbers = np.linspace(1 / wavelength_max, 1 / wavelength_min, 1001)
    wavelengths = 1 / wavenumbers  # mono mode setting
    bin_set = "1nm"  # ckd mode setting
    bins = "550"  # ckd mode setting

    # Determine artefact filename
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(
        outdir, f"test_onedim_ckd_vs_mono_{bins}_{reflectance}.png"
    )

    # Skip if data is missing, but still create an artefact
    for dataset_id in [
        "H2O-spectra-18000_18100",
        "H2O-spectra-18100_18200",
        "H2O-spectra-18200_18300",
        "H2O-spectra-18300_18400",
    ]:
        tu.skipif_data_not_found(
            f"spectra/absorption/{dataset_id}.nc",
            action=lambda: tu.missing_artefact(fname_plot),
        )

    # Monochromatic experiment
    mono_exp = init_mono_experiment(
        wavelengths=wavelengths,
        spp=spp,
        reflectance=reflectance,
        zeniths=zeniths,
    )
    mono_results = eradiate.run(mono_exp)

    wavelength_bin_width = (wavelengths.max() - wavelengths.min()).m_as(
        mono_results.w.attrs["units"]
    )
    mono_results_integrated = mono_results.integrate(coord="w") / wavelength_bin_width

    # CKD experiment
    ckd_exp = init_ckd_experiment(
        bin_set=bin_set,
        bins=bins,
        spp=spp,
        reflectance=reflectance,
        zeniths=zeniths,
    )
    ckd_results = eradiate.run(ckd_exp)

    # Make figure
    make_figure(
        ckd_results=ckd_results,
        mono_results_averaged=mono_results_integrated,
        fname_plot=fname_plot,
        wavelength_bin="[549.5, 550.5] nm",
        reflectance=reflectance,
    )

    # Assert averaged mono results are consistent with ckd results within 5 %
    mono_brf = mono_results_integrated.brf.squeeze().values
    ckd_brf = ckd_results.brf.squeeze().values
    assert np.allclose(ckd_brf, mono_brf, rtol=0.05)


@pytest.mark.parametrize("reflectance", [0.0, 0.5, 1.0])
@pytest.mark.slow
def test_1050(reflectance, artefact_dir):
    r"""
    Result consistency between ``mono_double`` and ``ckd_double`` modes
    ===================================================================

    This test checks the results consistency between mono_double and ckd_double
    modes in the [1049.5, 1050.5] nm wavelength bin.

    Rationale
    ---------

    * Sensor: Distant measure covering a plane (11 angular points,
      10000 sample per pixel) and targeting (0, 0, 0).
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta = 50.0°`.
    * Surface: a square surface with a Lambertian BRDF with reflectance
      parameter in :math:`\rho \in [0.0, 0.5, 1.0]`.
    * Atmosphere: a molecular atmosphere derived from the AFGL (1986)
      thermophysical properties ``us_standard`` model, with both scattering and
      absorption enabled.
    * Integrator: volumetric path tracer.

    Expected behaviour
    ------------------

    The [1049.5, 1050.5] nm - integrated BRF results obtained with the
    ``mono_double`` and ``ckd_double`` modes should agree within 5 %.

    The ``mono_double`` BRF results are spectrally averaged according to the
    following formula:

    .. math::

       \hat{q} = \frac{
          \int_{
             \lambda_{\mathrm{min}}
          }^{
             \lambda_{\mathrm{max}}
          }
          q(\lambda) \, \mathrm{d} \lambda
       }{
          \int_{
             \lambda_{\mathrm{min}}
          }^{
             \lambda_{\mathrm{max}}
          }
          \mathrm{d} \lambda
       }

    where

    * :math:`q` is some spectral quantity (e.g. the BRF)
    * :math:`\lambda` denotes the wavelength
    * :math:`[\lambda_{\mathrm{min}},\, \lambda_{\mathrm{max}}]` is the
      wavelength bin

    Results
    -------

    Note: :math:`\mathrm{RAD}(x, y) = \left| \dfrac{x - y}{x} \right|`

    .. image:: generated/plots/test_onedim_ckd_vs_mono_1050_0.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_ckd_vs_mono_1050_0.5.png
       :width: 95%
    .. image:: generated/plots/test_onedim_ckd_vs_mono_1050_1.0.png
       :width: 95%
    """
    # Settings
    zeniths = np.linspace(-75, 75, 11)
    spp = 1e4
    wavelength_min = 1049.5 * ureg.nm
    wavelength_max = 1050.5 * ureg.nm
    wavenumbers = np.linspace(1 / wavelength_max, 1 / wavelength_min, 1001)
    wavelengths = 1 / wavenumbers  # mono mode setting
    bin_set = "1nm"  # ckd mode setting
    bins = "1050"  # ckd mode setting

    # Determine artefact filename
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(
        outdir, f"test_onedim_ckd_vs_mono_{bins}_{reflectance}.png"
    )

    # Skip if data is missing, but still create an artefact
    for dataset_id in [
        "H2O-spectra-9400_9500",
        "H2O-spectra-9500_9600",
        "CO2-spectra-9400_9500",
        "CO2-spectra-9500_9600",
        "N2O-spectra-9400_9500",
        "N2O-spectra-9500_9600",
        "CO-spectra-9400_9500",
        "CO-spectra-9500_9600",
        "CH4-spectra-9400_9500",
        "CH4-spectra-9500_9600",
        "O2-spectra-9400_9500",
        "O2-spectra-9500_9600",
    ]:
        tu.skipif_data_not_found(
            f"spectra/absorption/{dataset_id}.nc",
            action=lambda: tu.missing_artefact(fname_plot),
        )

    # Monochromatic experiment
    mono_exp = init_mono_experiment(
        wavelengths=wavelengths,
        spp=spp,
        reflectance=reflectance,
        zeniths=zeniths,
    )
    mono_results = eradiate.run(mono_exp)

    wavelength_bin_width = (wavelengths.max() - wavelengths.min()).m_as(
        mono_results.w.attrs["units"]
    )
    mono_results_averaged = mono_results.integrate(coord="w") / wavelength_bin_width

    # CKD experiment
    ckd_exp = init_ckd_experiment(
        bin_set=bin_set,
        bins=bins,
        spp=spp,
        reflectance=reflectance,
        zeniths=zeniths,
    )
    ckd_results = eradiate.run(ckd_exp)

    # Make figure
    make_figure(
        ckd_results=ckd_results,
        mono_results_averaged=mono_results_averaged,
        fname_plot=fname_plot,
        wavelength_bin="[1049.5, 1050.5] nm",
        reflectance=reflectance,
    )

    # Assert averaged mono results are consistent with ckd results within 5 %
    mono_brf = mono_results_averaged.brf.squeeze().values
    ckd_brf = ckd_results.brf.squeeze().values
    assert np.allclose(ckd_brf, mono_brf, rtol=0.05)



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.thermoprops import us76

_Q = ureg.Quantity


def test_make_profile():
    # default constructor
    profile = us76.make_profile()
    assert profile["z_level"].values[0] == 0.0
    assert profile["z_level"].values[-1] == 100000.0
    assert profile.dims["z_layer"] == 50
    assert profile.dims["species"] == 12

    # custom levels altitudes
    profile = us76.make_profile(levels=_Q(np.linspace(2.0, 15.0, 51), "km"))
    assert profile.dims["z_layer"] == 50
    assert profile["z_level"].values[0] == 2000.0
    assert profile["z_level"].values[-1] == 15000.0
    assert profile.dims["species"] == 12

    # custom number of layers
    profile = us76.make_profile(levels=_Q(np.linspace(0.0, 150.0, 37), "kilometers"))
    assert profile.dims["z_layer"] == 36
    assert profile["z_level"].values[0] == 0.0
    assert profile["z_level"].values[-1] == 150000.0
    assert profile.dims["species"] == 12

    profile = us76.make_profile(levels=_Q(np.linspace(0.0, 80.0, 2), "kilometers"))
    assert profile.dims["z_layer"] == 1
    assert profile["z_level"].values[0] == 0.0
    assert profile["z_level"].values[-1] == 80000.0
    assert profile.dims["species"] == 12

    # invalid levels
    with pytest.raises(ValueError):
        us76.make_profile(levels=np.linspace(-4000, 50000))

    with pytest.raises(ValueError):
        us76.make_profile(levels=np.linspace(500.0, 5000000.0))


def test_create():
    z = _Q(np.linspace(0.0, 100000.0, 101), "meter")
    variables = ["p", "t", "n", "n_tot"]
    ds = us76.create(z, variables=variables)

    dims = ds.dims
    assert len(dims) == 2
    assert "z" in dims
    assert "species" in dims

    coords = ds.coords
    assert len(coords) == 2
    assert (coords["z"] == z.magnitude).all()
    assert [s for s in coords["species"]] == [s for s in us76.SPECIES]

    for var in variables:
        assert var in ds

    assert all(
        [
            x in ds.attrs
            for x in ["convention", "title", "history", "source", "references"]
        ]
    )


def test_create_below_86_km_layers_boundary_altitudes():
    """
    We test the computation of the atmospheric variables (pressure,
    temperature and mass density) at the level altitudes, i.e. at the model
    layer boundaries. We assert correctness by comparing their values with the
    values from the table 1 of the U.S. Standard Atmosphere 1976 document.
    """

    z = us76.to_altitude(np.array(us76.H))
    ds = us76.create(z, variables=["p", "t", "rho"])

    level_temperature = np.array(
        [288.15, 216.65, 216.65, 228.65, 270.65, 270.65, 214.65, 186.87]
    )
    level_pressure = np.array(
        [101325.0, 22632.0, 5474.8, 868.01, 110.90, 66.938, 3.9564, 0.37338]
    )
    level_mass_density = np.array(
        [
            1.225,
            0.36392,
            0.088035,
            0.013225,
            0.0014275,
            0.00086160,
            0.000064261,
            0.000006958,
        ]
    )

    assert np.allclose(ds["t"].values, level_temperature, rtol=1e-4)
    assert np.allclose(ds["p"].values, level_pressure, rtol=1e-4)
    assert np.allclose(ds["rho"].values, level_mass_density, rtol=1e-3)


def test_create_below_86_km_arbitrary_altitudes():
    """
    We test the computation of the atmospheric variables (pressure,
    temperature and mass density) at arbitrary altitudes. We assert correctness
    by comparing their values to the values from table 1 of the U.S. Standard
    Atmosphere 1976 document.
    """

    # The values below were selected arbitrarily from Table 1 of the document
    # such that there is at least one value in each of the 7 temperature
    # regions.
    h = np.array(
        [
            200.0,
            1450.0,
            5250.0,
            6500.0,
            9800.0,
            17900.0,
            24800.0,
            27100.0,
            37200.0,
            40000.0,
            49400.0,
            61500.0,
            79500.0,
            84000.0,
        ]
    )
    temperatures = np.array(
        [
            286.850,
            278.725,
            254.025,
            245.900,
            224.450,
            216.650,
            221.450,
            223.750,
            243.210,
            251.050,
            270.650,
            241.250,
            197.650,
            188.650,
        ]
    )
    pressures = np.array(
        [
            98945.0,
            85076.0,
            52239.0,
            44034.0,
            27255.0,
            7624.1,
            2589.6,
            1819.4,
            408.7,
            277.52,
            81.919,
            16.456,
            0.96649,
            0.43598,
        ]
    )
    mass_densities = np.array(
        [
            1.2017,
            1.0633,
            0.71641,
            0.62384,
            0.42304,
            0.12259,
            0.040739,
            0.028328,
            0.0058542,
            0.0038510,
            0.0010544,
            0.00023764,
            0.000017035,
            0.0000080510,
        ]
    )

    z = us76.to_altitude(h)
    ds = us76.create(z, variables=["t", "p", "rho"])

    assert np.allclose(temperatures, ds["t"].values, rtol=1e-4)
    assert np.allclose(pressures, ds["p"].values, rtol=1e-4)
    assert np.allclose(mass_densities, ds["rho"].values, rtol=1e-4)


def test_init_data_set():
    def check_data_set(ds):
        for var in us76.VARIABLES:
            assert var in ds
            assert np.isnan(ds[var].values).all()

        assert ds["n"].values.ndim == 2
        assert all(
            ds["species"].values
            == ["N2", "O2", "Ar", "CO2", "Ne", "He", "Kr", "Xe", "CH4", "H2", "O", "H"]
        )

    z1 = _Q(np.linspace(0.0, 50000.0), "meter")
    ds1 = us76.init_data_set(z1)
    check_data_set(ds1)

    z2 = _Q(np.linspace(120000.0, 650000.0), "meter")
    ds2 = us76.init_data_set(z2)
    check_data_set(ds2)

    z3 = _Q(np.linspace(70000.0, 100000.0), "meter")
    ds3 = us76.init_data_set(z3)
    check_data_set(ds3)


def test_compute_levels_temperature_and_pressure_low_altitude():
    tb, pb = us76.compute_levels_temperature_and_pressure_low_altitude()

    level_temperature = np.array(
        [288.15, 216.65, 216.65, 228.65, 270.65, 270.65, 214.65, 186.87]
    )
    level_pressure = np.array(
        [101325.0, 22632.0, 5474.8, 868.01, 110.90, 66.938, 3.9564, 0.37338]
    )

    assert np.allclose(tb, level_temperature, rtol=1e-3)
    assert np.allclose(pb, level_pressure, rtol=1e-3)


def rtol(v, ref):
    return np.abs(v - ref) / ref


def test_compute_number_density():
    # the following altitudes values are chosen arbitrarily
    altitudes = ureg.Quantity(
        np.array(
            [
                86.0,
                90.0,
                95.0,
                100.0,
                110.0,
                120.0,
                150.0,
                200.0,
                300.0,
                400.0,
                500.0,
                600.0,
                700.0,
                800.0,
                900.0,
                1000.0,
            ]
        ),
        "km",
    )
    mask = altitudes.magnitude > 150.0

    # the corresponding number density values are from NASA (1976) - U.S.
    # Standard Atmosphere, table VIII (p. 210-215)
    values = {
        "N2": np.array(
            [
                1.13e20,
                5.547e19,
                2.268e19,
                9.210e18,
                1.641e18,
                3.726e17,
                3.124e16,
                2.925e15,
                9.593e13,
                4.669e12,
                2.592e11,
                1.575e10,
                1.038e9,
                7.377e7,
                5.641e6,
                4.626e5,
            ]
        ),
        "O": np.array(
            [
                us76.O_7,
                2.443e17,
                4.365e17,
                4.298e17,
                2.303e17,
                9.275e16,
                1.780e16,
                4.050e15,
                5.443e14,
                9.584e13,
                1.836e13,
                3.707e12,
                7.840e11,
                1.732e11,
                3.989e10,
                9.562e9,
            ]
        ),
        "O2": np.array(
            [
                us76.O2_7,
                1.479e19,
                5.83e18,
                2.151e18,
                2.621e17,
                4.395e16,
                2.750e15,
                1.918e14,
                3.942e12,
                1.252e11,
                4.607e9,
                1.880e8,
                8.410e6,
                4.105e5,
                2.177e4,
                1.251e3,
            ]
        ),
        "Ar": np.array(
            [
                us76.AR_7,
                6.574e17,
                2.583e17,
                9.501e16,
                1.046e16,
                1.366e15,
                5.0e13,
                1.938e12,
                1.568e10,
                2.124e8,
                3.445e6,
                6.351e4,
                1.313e3,
                3.027e1,
                7.741e-1,
                2.188e-2,
            ]
        ),
        "He": np.array(
            [
                7.582e14,
                3.976e14,
                1.973e14,
                1.133e14,
                5.821e13,
                3.888e13,
                2.106e13,
                1.310e13,
                7.566e12,
                4.868e12,
                3.215e12,
                2.154e12,
                1.461e12,
                1.001e12,
                6.933e11,
                4.850e11,
            ]
        ),
        "H": np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                3.767e11,
                1.630e11,
                1.049e11,
                8.960e10,
                8.0e10,
                7.231e10,
                6.556e10,
                5.961e10,
                5.434e10,
                4.967e10,
            ]
        ),
    }

    n = us76.compute_number_densities_high_altitude(altitudes)

    # print('N2:', rtol(n[0], values['N2']))
    assert np.allclose(n["N2"], values["N2"], rtol=0.01)
    # print('O:',  rtol(n[1], values['O']))
    assert np.allclose(n["O"], values["O"], rtol=0.1)
    # print('O2:', rtol(n[2], values['O2']))
    assert np.allclose(n["O2"], values["O2"], rtol=0.01)
    # print('Ar:', rtol(n[3], values['Ar']))
    assert np.allclose(n["Ar"], values["Ar"], rtol=0.01)
    # print('He:', rtol(n[4], values['He']))
    assert np.allclose(n["He"], values["He"], rtol=0.01)
    # print('H:', rtol(n[5][mask], values['H'][mask]))
    assert np.allclose(n["H"][mask], values["H"][mask], rtol=0.01)


def test_compute_mean_molar_mass():
    # test call with scalar altitude
    assert us76.compute_mean_molar_mass_high_altitude(90.0) == us76.M0
    assert us76.compute_mean_molar_mass_high_altitude(200.0) == us76.M["N2"]

    # test call with array of altitudes
    z = np.linspace(86, 1000, 915)
    assert np.allclose(
        us76.compute_mean_molar_mass_high_altitude(z),
        np.where(z <= 100.0, us76.M0, us76.M["N2"]),
    )


def test_compute_temperature_above_86_km():
    # test altitudes out of range raises value error
    with pytest.raises(ValueError):
        us76.compute_temperature_high_altitude(10.0)

    # test call with scalar altitude
    assert np.isclose(us76.compute_temperature_high_altitude(90.0), 186.87, rtol=1e-3)

    # test call with array of altitudes
    z = [100, 110, 120, 130, 200, 500]  # km
    assert np.allclose(
        us76.compute_temperature_high_altitude(z),
        np.array([195.08, 240.00, 360.0, 469.27, 854.56, 999.24]),
        rtol=1e-3,
    )



from .ckd import Bin as Bin
from .ckd import BinSet as BinSet
from .index import CKDSpectralIndex as CKDSpectralIndex
from .index import MonoSpectralIndex as MonoSpectralIndex
from .index import SpectralIndex as SpectralIndex
from .mono import WavelengthSet as WavelengthSet



import pytest

import eradiate

# ------------------------------------------------------------------------------
#                            Pre-process helpers
# ------------------------------------------------------------------------------

pytest.register_assert_rewrite("eradiate.test_tools.types.check_scene_element")

# ------------------------------------------------------------------------------
#                                 Mode fixtures
# ------------------------------------------------------------------------------


def generate_fixture(mode):
    @pytest.fixture()
    def fixture():
        import eradiate

        eradiate.set_mode(mode)

    globals()["mode_" + mode] = fixture


for mode in eradiate.modes():
    generate_fixture(mode)
del generate_fixture


def generate_fixture_group(name, modes):
    @pytest.fixture(params=modes)
    def fixture(request):
        mode = request.param
        import eradiate

        eradiate.set_mode(mode)

    globals()["modes_" + name] = fixture


variants = [x for x in eradiate.modes() if x not in {"mono", "ckd"}]  # Remove aliases
variant_groups = {
    "all_mono": [x for x in variants if x.startswith("mono")],
    "all_ckd": [x for x in variants if x.startswith("ckd")],
    "all_mono_ckd": [
        x for x in variants if (x.startswith("mono") or x.startswith("ckd"))
    ],
    "all_single": [x for x in variants if x.endswith("single")],
    "all_double": [x for x in variants if x.endswith("double")],
    "all": variants,
}

for name, variants in variant_groups.items():
    generate_fixture_group(name, variants)
del generate_fixture_group


# ------------------------------------------------------------------------------
#                              Other configuration
# ------------------------------------------------------------------------------


@pytest.fixture
def ert_seed_state():
    from eradiate.rng import SeedState

    return SeedState(0)



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.radprops import US76ApproxRadProfile, ZGrid
from eradiate.spectral.index import SpectralIndex


@pytest.fixture
def absorption_dataset_550nm():
    return "spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"


def test_us76_approx_rad_profile(mode_mono, absorption_dataset_550nm):
    """
    Collision coefficient evaluation methods return pint.Quantity objects.
    """
    p = US76ApproxRadProfile(absorption_dataset=absorption_dataset_550nm)

    si = SpectralIndex.new()
    for field in ["sigma_a", "sigma_s", "sigma_t", "albedo"]:
        x = getattr(p, f"eval_{field}")(si, p.zgrid)
        assert isinstance(x, ureg.Quantity)


def test_us76_approx_rad_profile_levels(mode_mono, absorption_dataset_550nm):
    """
    Collision coefficients' shape match altitude levels shape.
    """
    p = US76ApproxRadProfile(
        thermoprops=dict(levels=ureg.Quantity(np.linspace(0, 120, 121), "km")),
        absorption_dataset=absorption_dataset_550nm,
    )

    si = SpectralIndex.new()
    for field in ["sigma_a", "sigma_s", "sigma_t", "albedo"]:
        x = getattr(p, f"eval_{field}")(si, p.zgrid)
        assert x.shape == (120,)


def test_us76_approx_rad_profile_has_absorption_default(mode_mono):
    """
    Default value for 'has_absorption' is True, hence the absorption
    coefficient is computed and is not zero everywhere at 1650 nm.
    """
    p = US76ApproxRadProfile(
        absorption_dataset="spectra/absorption/us76_u86_4/us76_u86_4-spectra-6000_7000.nc",
    )
    assert p.has_absorption
    si = SpectralIndex.new(w=1650 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_a.values != 0.0).any()


def test_us76_approx_rad_profile_has_absorption_true(mode_mono):
    """
    When 'has_absorption' is True, the absorption coefficient is computed
    and is not zero everywhere at 1650 nm.
    """
    p = US76ApproxRadProfile(
        absorption_dataset="spectra/absorption/us76_u86_4/us76_u86_4-spectra-6000_7000.nc",
        has_absorption=True,
    )
    assert p.has_absorption
    si = SpectralIndex.new(w=1650.0 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_a.values != 0.0).any()


def test_us76_approx_rad_profile_has_absorption_false(
    mode_mono,
    absorption_dataset_550nm,
):
    """
    When 'has_absorption' is False, the absorption coefficient is not
    computed and is zero everywhere.
    """
    p = US76ApproxRadProfile(
        absorption_dataset=absorption_dataset_550nm,
        has_absorption=False,
    )
    assert not p.has_absorption
    si = SpectralIndex.new(w=1650.0 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_a.values == 0.0).all()


def test_us76_approx_rad_profile_has_scattering_default(
    mode_mono,
    absorption_dataset_550nm,
):
    """
    Default value for 'has_scattering' is True, hence the scattering
    coefficient is computed and is not zero everywhere at 550 nm.
    """
    p = US76ApproxRadProfile(absorption_dataset=absorption_dataset_550nm)
    assert p.has_scattering
    si = SpectralIndex.new(w=550.0 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_s.values != 0.0).any()


def test_us76_approx_rad_profile_has_scattering_true(
    mode_mono,
    absorption_dataset_550nm,
):
    """
    When 'has_scattering' is True, the scattering coefficient is computed
    and is not zero everywhere at 550 nm.
    """
    p = US76ApproxRadProfile(
        has_scattering=True,
        absorption_dataset=absorption_dataset_550nm,
    )
    assert p.has_scattering
    si = SpectralIndex.new(w=550 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_s.values != 0.0).any()


def test_us76_approx_rad_profile_has_scattering_false(
    mode_mono,
    absorption_dataset_550nm,
):
    """
    When 'has_scattering' is False, the scattering coefficient is not
    computed and is zero everywhere.
    """
    p = US76ApproxRadProfile(
        has_scattering=False,
        absorption_dataset=absorption_dataset_550nm,
    )
    assert not p.has_scattering
    si = SpectralIndex.new(w=550.0 * ureg.nm)
    ds = p.eval_dataset(si, p.zgrid)
    assert (ds.sigma_s.values == 0.0).all()


def test_us76_approx_eval_zgrid(mode_mono, absorption_dataset_550nm):
    """
    Evaluation on an arbitrary altitude grid works.
    """
    p = US76ApproxRadProfile(absorption_dataset=absorption_dataset_550nm)
    si = SpectralIndex.new(w=550.0 * ureg.nm)
    zgrid = ZGrid(levels=np.linspace(0, 100, 1001) * ureg.km)
    ds = p.eval_dataset(si, zgrid)
    assert len(ds.z_layer) == len(zgrid.layers)



from __future__ import annotations

from pathlib import Path
from typing import cast

import typer
from click import Command, Group

import eradiate.cli


def get_docs_for_click_markdown(
    *,
    obj: Command,
    ctx: typer.Context,
    indent: int = 0,
    name: str = "",
    call_prefix: str = "",
) -> str:
    docs = "#" * (1 + indent)
    command_name = name or obj.name
    if call_prefix:
        command_name = f"{call_prefix} {command_name}"
    title = f"`{command_name}`" if command_name else "CLI"
    docs += f" {title}\n\n"
    if obj.help:
        docs += f"{obj.help}\n\n"
    usage_pieces = obj.collect_usage_pieces(ctx)
    if usage_pieces:
        docs += "**Usage**:\n\n"
        docs += "```console\n"
        docs += "$ "
        if command_name:
            docs += f"{command_name} "
        docs += f"{' '.join(usage_pieces)}\n"
        docs += "```\n\n"
    args = []
    opts = []
    for param in obj.get_params(ctx):
        rv = param.get_help_record(ctx)
        if rv is not None:
            if param.param_type_name == "argument":
                args.append(rv)
            elif param.param_type_name == "option":
                opts.append(rv)
    if args:
        docs += f"**Arguments**:\n\n"
        for arg_name, arg_help in args:
            docs += f"* `{arg_name}`"
            if arg_help:
                docs += f": {arg_help}"
            docs += "\n"
        docs += "\n"
    if opts:
        docs += f"**Options**:\n\n"
        for opt_name, opt_help in opts:
            docs += f"* `{opt_name}`"
            if opt_help:
                docs += f": {opt_help}"
            docs += "\n"
        docs += "\n"
    if obj.epilog:
        docs += f"{obj.epilog}\n\n"
    if isinstance(obj, Group):
        group: Group = cast(Group, obj)
        commands = group.list_commands(ctx)
        if commands:
            docs += f"**Commands**:\n\n"
            for command in commands:
                command_obj = group.get_command(ctx, command)
                assert command_obj
                docs += f"* `{command_obj.name}`"
                command_help = command_obj.get_short_help_str()
                if command_help:
                    docs += f": {command_help}"
                docs += "\n"
            docs += "\n"
        for command in commands:
            command_obj = group.get_command(ctx, command)
            assert command_obj
            use_prefix = ""
            if command_name:
                use_prefix += f"{command_name}"
            docs += get_docs_for_click_markdown(
                obj=command_obj, ctx=ctx, indent=indent + 1, call_prefix=use_prefix
            )
    return docs


def docs(typer_obj: typer.Typer, name: str = "", indent: int = 0) -> str:
    """
    Generate Markdown docs for a Typer app.
    Code adapted from the typer-cli project.
    """
    click_obj = typer.main.get_command(typer_obj)
    ctx = typer.Context(click_obj)
    docs = get_docs_for_click_markdown(obj=click_obj, ctx=ctx, name=name, indent=indent)
    clean_docs = f"{docs.strip()}\n"
    return clean_docs


def write_if_modified(filename, content):
    filename.parent.mkdir(parents=True, exist_ok=True)

    try:
        with open(filename, "r") as f:
            existing = f.read()
    except OSError:
        existing = None

    if existing == content:
        print(f"Skipping unchanged '{filename.name}'")

    else:
        print(f"Generating '{filename.name}'")
        with open(filename, "w") as f:
            f.write(content)


def generate():
    root_dir = Path(__file__).absolute().parent.parent
    out_dir = root_dir / "docs/src"
    print(f"Generating CLI docs in '{out_dir}'")

    docs_markdown = (
        "(sec-reference_cli)=\n"
        "# Command-line interface reference\n\n"
        "This is the reference for Eradiate’s command-line tools. It consists "
        "of a main entry point `eradiate`, and its multiple subcommands "
        "documented hereafter. The implementation is located in the "
        "`eradiate.cli` module (not documented here).\n\n"
        + docs(eradiate.cli.app, name="eradiate", indent=1)
    )
    write_if_modified(out_dir / "reference_cli.md", docs_markdown)


if __name__ == "__main__":
    generate()



from __future__ import annotations

import typing as t
from collections import abc as cabc

import attrs
import mitsuba as mi
import numpy as np

from ._core import PhaseFunction, phase_function_factory
from ..core import traverse
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ...attrs import documented
from ...contexts import KernelContext
from ...kernel import InitParameter, UpdateParameter, map_unit_cube
from ...spectral.index import SpectralIndex
from ...units import unit_context_kernel as uck
from ...util.misc import cache_by_id


@attrs.define(eq=False, slots=False)
class BlendPhaseFunction(PhaseFunction):
    """
    Blended phase function [``blend_phase``].

    This phase function aggregates two or more sub-phase functions
    (*components*) and blends them based on its `weights` parameter. Weights are
    usually based on the associated medium's scattering coefficient.
    """

    components: list[PhaseFunction] = documented(
        attrs.field(
            converter=lambda x: [phase_function_factory.convert(y) for y in x],
            validator=attrs.validators.deep_iterable(
                attrs.validators.instance_of(PhaseFunction)
            ),
            kw_only=True,
        ),
        type="list of :class:`.PhaseFunction`",
        init_type="list of :class:`.PhaseFunction` or list of dict",
        doc="List of components (at least two). This parameter has not default.",
    )

    @components.validator
    def _components_validator(self, attribute, value):
        if not len(value) > 1:
            raise ValueError(
                f"while validating {attribute.name}: BlendPhaseFunction must "
                "have at least two components"
            )

    weights: (np.ndarray | list[t.Callable[[KernelContext], np.ndarray]]) = documented(
        attrs.field(
            converter=lambda x: x if callable(x[0]) else np.array(x, dtype=np.float64),
            kw_only=True,
        ),
        type="ndarray or list of callables",
        init_type="array-like or list of callables",
        doc="List of weights associated with each component. Weights may be "
        "numerical values; in that case, they must be of shape (n,) or (n, m), "
        "where n is the number of components and m the number of cells along "
        "the atmosphere's vertical axis. Alternatively, weights may be "
        "callables that take a :class:`.KernelContext` as argument and "
        "return an array of shape (n, m). "
        "This parameter is required and has no default.",
    )

    @weights.validator
    def _weights_validator(self, attribute, value):
        if isinstance(value, np.ndarray):
            if value.ndim == 0 or value.ndim > 2:
                raise ValueError(
                    f"while validating '{attribute.name}': array must have 1 or 2 "
                    f"dimensions, got {value.ndim}"
                )

            if not value.shape[0] == len(self.components):
                raise ValueError(
                    f"while validating '{attribute.name}': array must have shape "
                    "(n,) or (n, m) where n is the number of components; got "
                    f"{value.shape}"
                )

        elif isinstance(value, cabc.Sequence):
            if not len(value) == len(self.components):
                raise ValueError(
                    f"while validating '{attribute.name}': weight and component "
                    "lists must have the same length"
                )

    geometry: SceneGeometry | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(SceneGeometry.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(SceneGeometry)
            ),
        ),
        doc="Parameters defining the basic geometry of the scene. If unset, "
        "the volume textures defining component weights will be assigned "
        "defaults likely unsuitable for atmosphere construction.",
        type=".SceneGeometry or None",
        init_type=".SceneGeometry or dict or str, optional",
        default="None",
    )

    def update(self) -> None:
        super().update()

        # Synchronize geometries
        for component in self.components:
            component.update()

            if isinstance(component, BlendPhaseFunction):
                component.geometry = self.geometry

    @cache_by_id
    def _eval_conditional_weights_impl(self, si: SpectralIndex) -> np.ndarray:
        """
        Memoised weight evaluation, used if weights are defined as callables.
        """
        n_comp = len(self.components)

        if isinstance(self.weights, list):
            weights = np.array([w(si) for w in self.weights], dtype=np.float64)
        else:  # if isinstance(self.weights, np.ndarray):
            weights = np.array(self.weights, dtype=np.float64)

        if weights.ndim < 2:
            weights = weights.reshape((-1, 1))

        result = np.zeros((n_comp - 1, *weights.shape[1:]), dtype=np.float64)

        # Compute conditional weights
        for i in range(n_comp - 1):
            # Normalize weights
            weights_sum = weights[i:, ...].sum(axis=0, keepdims=True)
            weights_normalized = np.divide(
                weights[i:, ...],
                weights_sum,
                where=weights_sum != 0.0,
                out=np.zeros_like(weights[i:, ...]),
            )
            # Aggregate weights of all components except the first one
            result[i] = weights_normalized[1:, ...].sum(axis=0, keepdims=True)

        return result

    def eval_conditional_weights(
        self,
        si: SpectralIndex,
        n_component: int | list[int] | None = None,
    ) -> np.ndarray:
        """
        Evaluate the conditional weights of specified Mitsuba phase function
        components.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral context.

        n_component : int or list of int, optional
            The index of the Mitsuba phase function component for which the
            conditional weight should be evaluated. If ``None``, the conditional
            weights of all components will be evaluated.

        Returns
        -------
        ndarray
            Conditional weights of the specified components as an array of shape
            (N, M) where n is the number of components and m the number of cells
            along the atmosphere's vertical axis.
        """
        if n_component is None:
            n_component = range(len(self.components) - 1)
        elif isinstance(n_component, int):
            n_component = [n_component]

        # Compute normalized component weights (cached until call with different context)
        weights = self._eval_conditional_weights_impl(si)

        # Return selected components
        return weights[n_component, ...]

    @property
    def template(self) -> dict:
        result = {"type": "blendphase"}

        for i in range(len(self.components) - 1):
            prefix = "phase_1." * i

            # Add components
            template, _ = traverse(self.components[i])
            result.update(
                {
                    **{f"{prefix}phase_0.{k}": v for k, v in template.items()},
                    f"{prefix}phase_1.type": "blendphase",
                }
            )

            # Assign conditional weight to second component
            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                # Note: This defines a partial and evaluates the component index.
                # Passing i as the kwarg default value is essential to force the
                # dereferencing of the loop variable.
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return mi.VolumeGrid(
                        np.reshape(
                            self.eval_conditional_weights(ctx.si, n_component),
                            (-1, 1, 1),  # Mind dim ordering! (C-style, i.e. zyx)
                        ).astype(np.float32)
                    )

                result[f"{prefix}weight.type"] = "gridvolume"
                result[f"{prefix}weight.grid"] = InitParameter(eval_conditional_weights)

                if self.geometry is not None:
                    result[
                        f"{prefix}weight.to_world"
                    ] = self.geometry.atmosphere_volume_to_world

            elif isinstance(self.geometry, SphericalShellGeometry):
                # Same comment as above
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return mi.VolumeGrid(
                        np.reshape(
                            self.eval_conditional_weights(ctx.si, n_component),
                            (1, 1, -1),  # Mind dim ordering! (C-style, i.e. zyx)
                        ).astype(np.float32)
                    )

                result[f"{prefix}weight.type"] = "sphericalcoordsvolume"
                result[f"{prefix}weight.volume.type"] = "gridvolume"
                result[f"{prefix}weight.volume.grid"] = InitParameter(
                    eval_conditional_weights
                )
                result[
                    f"{prefix}weight.to_world"
                ] = self.geometry.atmosphere_volume_to_world
                result[f"{prefix}weight.rmin"] = self.geometry.atmosphere_volume_rmin

            else:
                raise ValueError(
                    f"unhandled scene geometry type '{type(self.geometry).__name__}'"
                )

        else:
            template, _ = traverse(self.components[-1])
            result.update({**{f"{prefix}phase_1.{k}": v for k, v in template.items()}})

        return result

    @property
    def params(self) -> dict[str, UpdateParameter]:
        result = {}

        for i in range(len(self.components) - 1):
            prefix = "phase_1." * i

            # Add components
            _, params = traverse(self.components[i])
            result.update(
                {
                    **{f"{prefix}phase_0.{k}": v for k, v in params.items()},
                }
            )

            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                # Note: This defines a partial and evaluates the component index.
                # Passing i as the kwarg default value is essential to force the
                # dereferencing of the loop variable.
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return np.reshape(
                        self.eval_conditional_weights(ctx.si, n_component),
                        (-1, 1, 1, 1),  # Mind dim ordering! (C-style, i.e. zyxc)
                    ).astype(np.float32)

                # Assign conditional weight to second component
                result[f"{prefix}weight.data"] = UpdateParameter(
                    eval_conditional_weights,
                    UpdateParameter.Flags.SPECTRAL,
                )

            elif isinstance(self.geometry, SphericalShellGeometry):
                # Same comment as above
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return np.reshape(
                        self.eval_conditional_weights(ctx.si, n_component),
                        (1, 1, -1, 1),  # Mind dim ordering! (C-style, i.e. zyxc)
                    ).astype(np.float32)

                # Assign conditional weight to second component
                result[f"{prefix}weight.volume.data"] = UpdateParameter(
                    eval_conditional_weights,
                    UpdateParameter.Flags.SPECTRAL,
                )

            else:
                raise NotImplementedError

        else:
            _, params = traverse(self.components[-1])
            result.update({**{f"{prefix}phase_1.{k}": v for k, v in params.items()}})

        return result



"""
AFGL (1986) radiative profile definition.
"""

from __future__ import annotations

import typing as t
import warnings
from functools import partial

import attrs
import numpy as np
import pandas as pd
import pint
import xarray as xr

from ._core import RadProfile, ZGrid, make_dataset
from .rayleigh import compute_sigma_s_air
from .. import converters
from ..attrs import documented, parse_docs
from ..spectral.ckd import G16
from ..thermoprops import afgl_1986
from ..thermoprops.util import (
    column_mass_density,
    column_number_density,
    volume_mixing_ratio_at_surface,
)
from ..units import to_quantity
from ..units import unit_registry as ureg
from ..util.misc import cache_by_id, summary_repr


def _convert_thermoprops_afgl_1986(value: t.MutableMapping | xr.Dataset) -> xr.Dataset:
    if isinstance(value, dict):
        return afgl_1986.make_profile(**value)
    else:
        return value


@parse_docs
@attrs.define(eq=False)
class AFGL1986RadProfile(RadProfile):
    """
    Radiative properties profile corresponding to the AFGL (1986) atmospheric
    thermophysical properties profiles
    :cite:`Anderson1986AtmosphericConstituentProfiles`.

    Warnings
    --------
    This class does not support ``mono`` modes.
    """

    absorption_dataset: xr.Dataset = documented(
        attrs.field(
            converter=converters.to_dataset(
                load_from_id=None  # TODO: open dataset (not load)
            ),
            validator=attrs.validators.instance_of(xr.Dataset),
        ),
        doc="Absorption coefficient dataset.",
        type="Dataset",
        init_type="Dataset or :class:`.PathLike`",
    )

    _thermoprops: xr.Dataset = documented(
        attrs.field(
            factory=lambda: afgl_1986.make_profile(),
            converter=_convert_thermoprops_afgl_1986,
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Thermophysical properties.",
        type="Dataset",
        default=":func:`~eradiate.thermoprops.afgl_1986.make_profile`",
    )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    _zgrid: ZGrid | None = attrs.field(default=None, init=False)

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        self._zgrid = ZGrid(levels=self.levels)

    @property
    def thermoprops(self) -> xr.Dataset:
        return self._thermoprops

    @property
    def levels(self) -> pint.Quantity:
        return to_quantity(self.thermoprops.z_level)

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @cache_by_id
    def _thermoprops_interp(self, zgrid: ZGrid) -> xr.Dataset:
        # Interpolate thermophysical profile on specified altitude grid
        # Note: we use a nearest neighbour scheme (so far, it doesn't seem to make a difference)
        # Note: this value is cached so that repeated calls with the same zgrid
        #       won't trigger an unnecessary computation.
        with xr.set_options(keep_attrs=True):
            result = self.thermoprops.interp(
                z_layer=zgrid.layers.m_as(self.thermoprops.z_layer.units),
                method="nearest",
                kwargs={"fill_value": "extrapolate"},
            )

        return result.assign_coords(
            z_level=(
                "z_level",
                zgrid.levels.m_as(self.thermoprops.z_level.units),
                self.thermoprops.z_level.attrs,
            )
        )

    def eval_albedo_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        sigma_t = self.eval_sigma_t_ckd(w=w, g=g, zgrid=zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_sigma_a_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        thermoprops = self._thermoprops_interp(zgrid)

        if not self.has_absorption:
            return ureg.Quantity(np.zeros(thermoprops.z_layer.size), "km^-1")

        ds = self.absorption_dataset

        # Combine the 'bin' and 'index' coordinates into a multi-index, then reindex dataset
        idx = pd.MultiIndex.from_arrays(
            (ds.bin.values, ds.index.values), names=("bin", "index")
        )
        ds = ds.drop_vars(("bin", "index"))
        ds = ds.reindex({"bd": idx.values})

        # This table maps species to the function used to compute
        # corresponding physical quantities used for concentration rescaling
        compute = {
            "H2O": partial(column_mass_density, species="H2O"),
            "CO2": partial(volume_mixing_ratio_at_surface, species="CO2"),
            "O3": partial(column_number_density, species="O3"),
        }

        # Evaluate species concentrations
        # Note: This includes a conversion to units appropriate for interpolation
        rescaled_species = list(compute.keys())
        concentrations = {
            species: compute[species](ds=thermoprops).m_as(ds[species].units)
            for species in rescaled_species
        }

        # Convert altitude to units appropriate for interpolation
        z = ureg.convert(thermoprops.z_layer, thermoprops.z_layer.units, ds.z.units)

        # Interpolate absorption coefficient on concentration coordinates
        bin_id = str(int(w.m_as("nm")))
        g_index = np.abs(g - G16).argmin()  # TODO: PR#311 hack

        sigma_a_value = (
            ds.k.sel(bd=(bin_id, g_index))
            .interp(
                z=z,
                kwargs=dict(
                    fill_value=0.0
                ),  # extrapolate to 0.0 for altitude out of bounds
            )
            .interp(
                **concentrations,
                kwargs=dict(
                    bounds_error=True  # raise when concentration are out of bounds
                ),
            )
            .values
        )

        return ureg.Quantity(sigma_a_value, ds.k.units)

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        thermoprops = self._thermoprops_interp(zgrid)

        if self.has_scattering:
            return compute_sigma_s_air(
                wavelength=w,
                number_density=to_quantity(thermoprops.n),
            )
        else:
            return ureg.Quantity(np.zeros((1, thermoprops.z_layer.size)), "km^-1")

    def eval_sigma_s_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w=w, zgrid=zgrid)

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_dataset_ckd(self, w: pint.Quantity, g: float, zgrid: ZGrid) -> xr.Dataset:
        thermoprops = self._thermoprops_interp(zgrid)

        return make_dataset(
            wavelength=w,
            z_level=to_quantity(thermoprops.z_level),
            z_layer=to_quantity(thermoprops.z_layer),
            sigma_a=self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid),
            sigma_s=self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid),
        ).squeeze()



from importlib.metadata import PackageNotFoundError, version

# Internal constants
_EXPECTED_MITSUBA_VERSION = "3.2.1"
_EXPECTED_MITSUBA_PATCH_VERSION = "0.0.2"

# Exported constants
ERADIATE_MITSUBA_PACKAGE_VERSION = None
ERADIATE_MITSUBA_PACKAGE = False
ERADIATE_KERNEL_VERSION = None
ERADIATE_KERNEL_PATCH_VERSION = None

# Check if Dr.Jit and Mitsuba can be imported successfully
try:
    __import__("drjit")
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError(
        "Could not import module 'drjit'; did you build the kernel and add "
        "it to your $PYTHONPATH?"
    ) from e

try:
    _mi = __import__("mitsuba")
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError(
        "Could not import module 'mitsuba'; did you build the kernel and add "
        "it to your $PYTHONPATH?"
    ) from e

# Check if the eradiate-mitsuba package is installed
try:
    ERADIATE_MITSUBA_PACKAGE_VERSION = version("eradiate-mitsuba")
    ERADIATE_MITSUBA_PACKAGE = True
except PackageNotFoundError as pkg_error:
    pass

# Retrieve the kernel and kernel patch versions
ERADIATE_KERNEL_VERSION = _mi.scalar_mono.MI_VERSION
ERADIATE_KERNEL_PATCH_VERSION = _mi.scalar_rgb.ERD_MI_VERSION

# Check if the kernel version is compatible
if ERADIATE_KERNEL_VERSION != _EXPECTED_MITSUBA_VERSION:
    raise ImportError(
        "Using an incompatible version of Mitsuba. Eradiate requires Mitsuba "
        f"{_EXPECTED_MITSUBA_VERSION}. Found Mitsuba {ERADIATE_KERNEL_VERSION}."
    )
if ERADIATE_KERNEL_PATCH_VERSION != _EXPECTED_MITSUBA_PATCH_VERSION:
    raise ImportError(
        "Using an incompatible patch version of Mitsuba. Eradiate requires a "
        f"Mitsuba kernel version {_EXPECTED_MITSUBA_VERSION}, with the patch "
        f"version {_EXPECTED_MITSUBA_PATCH_VERSION}. Found Mitsuba "
        f"{ERADIATE_KERNEL_VERSION} with the patch version"
        f" {ERADIATE_KERNEL_PATCH_VERSION}."
    )

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



import numpy as np
import pint
import pytest

from eradiate import unit_registry as ureg
from eradiate.radprops import AFGL1986RadProfile, ZGrid
from eradiate.spectral.index import SpectralIndex


@pytest.fixture
def test_ckd_si_1650():
    return SpectralIndex.new(w=1650 * ureg.nm)


@pytest.fixture
def test_ckd_si_550():
    return SpectralIndex.new(w=550.0 * ureg.nm)


@pytest.mark.parametrize("model_id", ["midlatitude_summer", "us_standard"])
def test_afgl_1986_rad_profile_default_ckd(mode_ckd, model_id):
    """
    Collision coefficient evaluation methods return pint.Quantity objects.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset=f"ckd/absorption/10nm/afgl_1986-{model_id}-10nm-v3.nc",
    )

    si = SpectralIndex.new()
    for field in ["albedo", "sigma_a", "sigma_s", "sigma_t"]:
        x = getattr(profile, f"eval_{field}")(si, profile.zgrid)
        assert isinstance(x, ureg.Quantity)


@pytest.mark.parametrize("model_id", ["midlatitude_summer", "us_standard"])
def test_afgl_1986_rad_profile_has_absorption_default(
    mode_ckd, test_ckd_si_1650, model_id
):
    """
    Default value for 'has_absorption' is True, hence the absorption
    coefficient is computed and is not zero everywhere at 1650 nm.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset=f"ckd/absorption/10nm/afgl_1986-{model_id}-10nm-v3.nc",
    )
    assert profile.has_absorption
    ds = profile.eval_dataset(test_ckd_si_1650, profile.zgrid)
    assert (ds.sigma_a.values != 0.0).any()


@pytest.mark.parametrize("model_id", ["midlatitude_summer", "us_standard"])
def test_afgl_1986_rad_profile_has_absorption_true(
    mode_ckd, test_ckd_si_1650, model_id
):
    """
    When 'has_absorption' is True, the absorption coefficient is computed
    and is not zero everywhere at 1650 nm.
    """
    profile = AFGL1986RadProfile(
        thermoprops=dict(model_id=model_id),
        has_absorption=True,
        absorption_dataset=f"ckd/absorption/10nm/afgl_1986-{model_id}-10nm-v3.nc",
    )
    assert profile.has_absorption
    ds = profile.eval_dataset(test_ckd_si_1650, profile.zgrid)
    assert (ds.sigma_a.values != 0.0).any()


@pytest.mark.parametrize("model_id", ["midlatitude_summer", "us_standard"])
def test_afgl_1986_rad_profile_has_absorption_false(
    mode_ckd, test_ckd_si_1650, model_id
):
    """
    When 'has_absorption' is False, the absorption coefficient is not
    computed and is zero everywhere.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset=f"ckd/absorption/10nm/afgl_1986-{model_id}-10nm-v3.nc",
        thermoprops=dict(model_id=model_id),
        has_absorption=False,
    )
    assert not profile.has_absorption
    ds = profile.eval_dataset(test_ckd_si_1650, profile.zgrid)
    assert (ds.sigma_a.values == 0.0).all()


def test_afgl_1986_rad_profile_has_scattering_default(mode_ckd, test_ckd_si_550):
    """
    Default value for 'has_scattering' is True, hence the absorption
    coefficient is computed and is not zero everywhere at 550 nm.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc",
    )
    assert profile.has_scattering
    ds = profile.eval_dataset(test_ckd_si_550, profile.zgrid)
    assert (ds.sigma_s.values != 0.0).any()


def test_afgl_1986_rad_profile_has_scattering_true(mode_ckd, test_ckd_si_550):
    """
    When 'has_scattering' is True, the scattering coefficient is computed
    and is not zero everywhere at 550 nm.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc",
        has_scattering=True,
    )
    assert profile.has_scattering
    ds = profile.eval_dataset(test_ckd_si_550, profile.zgrid)
    assert (ds.sigma_s.values != 0.0).any()


def test_afgl_1986_rad_profile_has_scattering_false(mode_ckd, test_ckd_si_550):
    """
    When 'has_scattering' is False, the scattering coefficient is not
    computed and is zero everywhere.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc",
        has_scattering=False,
    )
    assert not profile.has_scattering
    ds = profile.eval_dataset(test_ckd_si_550, profile.zgrid)
    assert (ds.sigma_s.values == 0.0).all()


@pytest.mark.parametrize(
    "model_id",
    [
        "tropical",
        "midlatitude_summer",
        "midlatitude_winter",
        "subarctic_summer",
        "subarctic_winter",
        "us_standard",
    ],
)
def test_afgl_1986_rad_profile_model_id(mode_ckd, model_id):
    """
    All models are supported in ckd mode.
    """
    AFGL1986RadProfile(
        absorption_dataset=f"ckd/absorption/10nm/afgl_1986-{model_id}-10nm-v3.nc",
        thermoprops=dict(model_id=model_id),
    )


@pytest.mark.parametrize(
    "molecule",
    ["N2O", "CO", "CH4", "O2"],
)
def test_afgl_1986_rad_profile_concentrations_ckd_not_implemented(mode_ckd, molecule):
    """
    Concentrations rescaling is not implemented for molecules other than
    H2O and O3 in CKD mode.
    """
    with pytest.raises(NotImplementedError):
        AFGL1986RadProfile(
            absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc",
            thermoprops=dict(concentrations={molecule: 0.0 * ureg.dimensionless}),
        )


@pytest.mark.parametrize(
    "w", [280.0, 550.0, 790.0, 1040.0, 1270.0, 1590.0, 2220.0, 2400.0] * ureg.nm
)
@pytest.mark.parametrize("model_id", ["midlatitude_summer", "us_standard"])
def test_afgl_1986_rad_profile_ckd_10nm(mode_ckd, w, model_id):
    """
    Can evaluate absorption coefficient.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc",
        thermoprops=dict(model_id=model_id),
    )
    si = SpectralIndex.new(w=w)
    sigma_a = profile.eval_sigma_a(si, profile.zgrid)
    assert isinstance(sigma_a, pint.Quantity)


def test_afgl_1986_eval_zgrid(mode_ckd):
    """
    Evaluation on an arbitrary altitude grid works.
    """
    profile = AFGL1986RadProfile(
        absorption_dataset="ckd/absorption/10nm/afgl_1986-us_standard-10nm-v3.nc"
    )
    si = SpectralIndex.new()
    zgrid = ZGrid(levels=np.linspace(0, 100, 1001) * ureg.km)
    ds = profile.eval_dataset(si, zgrid)
    assert len(ds.z_layer) == len(zgrid.layers)



from __future__ import annotations

import typing as t
import warnings

import attrs
import numpy as np
import pint
import xarray as xr

from ._core import RadProfile, ZGrid, make_dataset
from .absorption import compute_sigma_a
from .rayleigh import compute_sigma_s_air
from .. import converters
from ..attrs import documented, parse_docs
from ..thermoprops import us76
from ..units import to_quantity
from ..units import unit_registry as ureg
from ..util.misc import cache_by_id, summary_repr


def _convert_thermoprops_us76_approx(
    value: t.MutableMapping | xr.Dataset,
) -> xr.Dataset:
    if isinstance(value, dict):
        return us76.make_profile(**value)
    else:
        return value


@parse_docs
@attrs.define(eq=False)
class US76ApproxRadProfile(RadProfile):
    """
    Radiative properties profile approximately corresponding to an
    atmospheric profile based on the original U.S. Standard Atmosphere 1976
    atmosphere model.

    Warnings
    --------
    This class does not support ``ckd`` modes.

    Notes
    -----
    * The :mod:`~eradiate.thermoprops.us76` module implements the original *U.S.
      Standard Atmosphere 1976* atmosphere model, as defined by the
      :cite:`NASA1976USStandardAtmosphere` technical report.
      In the original atmosphere model, the gases are assumed well-mixed below
      the altitude of 86 kilometers.
      In the present radiative properties profile, the absorption coefficient is
      computed using the ``spectra-us76_u86_4`` absorption dataset.
      This dataset provides the absorption cross-section of a specific mixture
      of N2, O2, CO2 and CH4, the mixing ratio of which are those defined by the
      *U.S. Standard Atmosphere 1976* model for the region of altitudes under
      86 kilometers, where these four gas species are well-mixed.
      As a result, the dataset is representative of the *U.S. Standard Atmosphere
      1976* model only below 86 kilometers.
      Since the atmosphere is typically a hundred kilometers high or more in
      radiative transfer applications, and in order to make the radiative
      properties profile reach these altitudes, the absorption coefficient
      is nevertheless computed using the ``spectra-us76_u86_4`` dataset.
      This approximation assumes that the absorption coefficient does not vary
      much whether the mixing ratios of the absorbing gas mixture are those
      below or above 86 km.

    * Furthermore, the *U.S. Standard Atmosphere 1976* model includes other gas
      species than N2, O2, CO2 and CH4.
      They are: Ar, He, Ne, Kr, H, O, Xe, He and H2.
      All these species except H2 are absent from the
      `HITRAN <https://hitran.org/>`_ spectroscopic database.
      Since the absorption datasets are computed using HITRAN, the atomic species
      could not be included in ``spectra-us76_u86_4``.
      H2 was mistakenly forgotten and should be added to the dataset in a future
      revision.

    * We refer to the *U.S. Standard Atmosphere 1976* atmosphere model as the
      model defined by the set of assumptions and equations in part 1 of the
      report, and "numerically" illustrated by the extensive tables in part
      4 of the report.
      In particular, the part 3, entitled *Trace constituents*, which
      provides rough estimates and discussions on the amounts of trace
      constituents such as ozone, water vapor, nitrous oxide, methane, and so
      on, is not considered as part of the *U.S. Standard Atmosphere 1976*
      atmosphere model because it does not clearly define the concentration
      values of all trace constituents at all altitudes, neither does it
      provide a way to compute them.

    * It seems that the identifier "US76" is commonly used to refer to a
      standard atmospheric profile used in radiative transfer applications.
      However, there appears to be some confusion around the definition of
      that standard atmospheric profile.
      In our understanding, what is called the "US76 standard atmospheric
      profile", or "US76" in short, **is not the U.S. Standard Atmosphere
      1976 atmosphere model** but instead the so-called "U.S. Standard (1976)
      atmospheric constituent profile model" in a AFGL technical report
      entitled *AFGL Atmospheric Constituent Profiles (0-120km)* and
      published in 1986 by Anderson et al.
      :cite:`Anderson1986AtmosphericConstituentProfiles`.
      Although the "U.S. Standard (1976) atmospheric profile model" of the
      AFGL's report is based on the *U.S. Standard Atmosphere* 1976 atmosphere
      model (hence the name), it is significantly different when it comes
      about the gas species concentration profiles.
      Notably, the "U.S. Standard (1976) atmospheric profile model" of the
      AFGL's report include radiatively active gases such as H2O, O3, N2O,
      and CO, that the *U.S. Standard Atmosphere 1976* atmosphere model does
      not include.
    """

    absorption_dataset: xr.Dataset = documented(
        attrs.field(
            converter=converters.to_dataset(load_from_id=None),
            validator=attrs.validators.instance_of(xr.Dataset),
        ),
        doc="Absorption coefficient data set.",
        type=":class:`xarray.Dataset`",
    )

    _thermoprops: xr.Dataset = documented(
        attrs.field(
            factory=lambda: us76.make_profile(),
            converter=_convert_thermoprops_us76_approx,
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Thermophysical properties.",
        type="Dataset",
        default="us76.make_profile",
    )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    _zgrid: ZGrid | None = attrs.field(default=None, init=False)

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        self._zgrid = ZGrid(levels=self.levels)

    @property
    def thermoprops(self) -> xr.Dataset:
        """
        Return thermophysical properties.
        """
        return self._thermoprops

    @property
    def levels(self) -> pint.Quantity:
        """
        Return level altitudes.
        """
        return to_quantity(self.thermoprops.z_level)

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @cache_by_id
    def _thermoprops_interp(self, zgrid: ZGrid) -> xr.Dataset:
        # Interpolate thermophysical profile on specified altitude grid
        # Note: we use a nearest neighbour scheme
        # Note: this value is cached so that repeated calls with the same zgrid
        #       won't trigger an unnecessary computation.
        with xr.set_options(keep_attrs=True):
            result = self.thermoprops.interp(
                z_layer=zgrid.layers.m_as(self.thermoprops.z_layer.units),
                method="nearest",
                kwargs={"fill_value": "extrapolate"},
            )

        return result.assign_coords(
            z_level=(
                "z_level",
                zgrid.levels.m_as(self.thermoprops.z_level.units),
                self.thermoprops.z_level.attrs,
            )
        )

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Inherit docstring
        sigma_s = self.eval_sigma_s_mono(w, zgrid)
        sigma_t = self.eval_sigma_t_mono(w, zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_a_mono(w, zgrid) + self.eval_sigma_s_mono(w, zgrid)

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        profile = self._thermoprops_interp(zgrid)

        if self.has_absorption:
            ds = self.absorption_dataset

            # Compute scattering coefficient
            result = compute_sigma_a(
                ds=ds,
                wl=w,
                p=to_quantity(profile.p),
                n=to_quantity(profile.n),
                fill_values=dict(
                    pt=0.0
                ),  # us76_u86_4 dataset is limited to pressures above
                # 0.101325 Pa, but us76 thermophysical profile goes below that
                # value for altitudes larger than 93 km. At these altitudes, the
                # number density is so small compared to that at the sea level that
                # we assume it is negligible.
            )
            ds.close()
            return result
        else:
            return np.zeros(profile.z_layer.size) * ureg.km**-1

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        profile = self._thermoprops_interp(zgrid)

        if self.has_scattering:
            return compute_sigma_s_air(
                wavelength=w,
                number_density=to_quantity(profile.n),
            )

        else:
            return np.zeros(profile.z_layer.size) * ureg.km**-1

    def eval_dataset_mono(self, w: pint.Quantity, zgrid: ZGrid) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_mono(w, zgrid),
            sigma_s=self.eval_sigma_s_mono(w, zgrid),
        ).squeeze()



"""
Utility functions to manipulate atmosphere thermophysical properties data
sets.
"""
from __future__ import annotations

import warnings
from datetime import datetime

import numpy as np
import pint
import scipy.constants
import xarray as xr

from .. import data
from ..units import to_quantity
from ..units import unit_registry as ureg

ATOMIC_MASS_CONSTANT = ureg.Quantity(
    *scipy.constants.physical_constants["atomic mass constant"][:-1]
)


def column_number_density(ds: xr.Dataset, species: str) -> pint.Quantity:
    """
    Computes the column number density of a given species in an atmospheric
    profile.
    The column number density is computed according to the formula:

    .. math::
       N = \\sum_{i=0}^{L-1} n_i \\, (z_{i+1} - z_i)

    where

    :math:`N` denotes the column number density,
    :math:`z_i` are the level altitudes and
    :math:`n_i` are the number densities of that given species inside the
    :math:`L` atmospheric layers.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical properties data set.

    species : str
        Species.

    Returns
    -------
    quantity
        Column number density.
    """
    mr = ds.mr.sel(species=species).values
    n = to_quantity(ds.n)
    n_species = mr * n
    z_level = to_quantity(ds.z_level)
    dz = z_level[1:] - z_level[:-1]
    return (n_species * dz).sum()


def column_mass_density(ds: xr.Dataset, species: str) -> pint.Quantity:
    """
    Computes the column (mass) density of a given species in an atmospheric
    profile.
    The column mass density is computed according to the formula:

    .. math::
       \\sigma = N * m_u

    where

    :math:`\\sigma` denotes the column mass density,
    :math:`N` is the column number density,
    :math:`m_u` is the atomic mass constant.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical property data set.

    species : str
        Species.

    Returns
    -------
    quantity
        Column mass density.
    """
    with data.open_dataset("chemistry/molecular_masses.nc") as molecular_mass:
        m = molecular_mass.m.sel(s=species).values * ATOMIC_MASS_CONSTANT

    return m * column_number_density(ds=ds, species=species)


def number_density_at_surface(ds: xr.Dataset, species: str) -> pint.Quantity:
    """
    Compute the number density at the surface of a given species in an
    atmospheric profile.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical properties data set.

    species : str
        Species.

    Returns
    -------
    :class:`~pint.Quantity`
        Number density at the surface.
    """
    surface_mr = to_quantity(ds.mr.sel(species=species))[0]
    surface_n = to_quantity(ds.n)[0]
    return surface_mr * surface_n


def mass_density_at_surface(ds: xr.Dataset, species: str) -> pint.Quantity:
    """Compute the mass density at the surface of a given species in an
    atmospheric profile.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical properties data set.

    species : str
        Species.

    Returns
    -------
    quantity
        Mass density at the surface.
    """
    with data.open_dataset("chemistry/molecular_masses.nc") as molecular_mass:
        m = to_quantity(molecular_mass.m.sel(s=species)) * ATOMIC_MASS_CONSTANT

    return m * number_density_at_surface(ds=ds, species=species)


def volume_mixing_ratio_at_surface(ds: xr.Dataset, species: str) -> pint.Quantity:
    """Compute the volume mixing ratio at the surface of a given species in an
    atmospheric profile.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical properties data set.

    species : str
        Species.

    Returns
    -------
    quantity
        Volume mixing ratio at the surface.
    """
    return to_quantity(ds.mr.sel(species=species))[0]


def _scaling_factor(
    initial_amount: pint.Quantity,
    target_amount: pint.Quantity,
) -> float:
    """
    Compute the scaling factor to reach target amount from initial amount.

    Initial and target amount should be scalar and  have compatible dimensions.

    Parameters
    ----------
    initial_amount: quantity
        Initial amount.

    target_amount: quantity
        Target quantity.

    Raises
    ------
    ValueError
        When both initial and target amounts are zero.

    Returns
    -------
    float
        Scaling factor.
    """
    if initial_amount.m == 0.0:
        if target_amount.m == 0.0:
            return 0.0
        else:
            raise ValueError(
                f"Cannot compute scaling factor when initial amount is 0.0 and "
                f"target amount is non-zero (got {target_amount})"
            )
    else:
        return (target_amount / initial_amount).m_as(ureg.dimensionless)


def compute_scaling_factors(
    ds: xr.Dataset, concentration: dict[str, pint.Quantity]
) -> dict[str, float]:
    r"""
    Compute the scaling factors to be applied to the mixing ratio values
    of each species in an atmosphere thermophysical properties data set, so
    that the integrated number/mass density and/or the number/mass density at
    the surface, match given values.

    Parameters
    ----------
    ds : Dataset
        Atmosphere thermophysical properties data set.

    concentration : dict
        Mapping of species (str) and target concentration
        (:class:`~pint.Quantity`).

        If the target concentration has dimensions of inverse square length
        (:math:`[L^{-2}]`), the value is interpreted as a column
        number density for that given species and the scaling factor, :math:`f`,
        is obtained by dividing that column number density,
        :math:`N_{\mathrm{target}}`,
        by the initial column number density,
        :math:`N_{\mathrm{initial}}`:

        .. math::
           f = \frac{N_{\mathrm{target}}}{N_{\mathrm{initial}}}

        If the target concentration has dimensions of mass times inverse square
        length (:math:`[ML^{-2}]`), the value is interpreted as a column (mass)
        density for that species and the scaling factor is obtained by dividing
        that column mass density,
        :math:`\sigma_{\mathrm{target}}`,
        by the initial column mass density,
        :math:`\sigma_{\mathrm{initial}}`:

        .. math::
           f = \frac{\sigma_{\mathrm{target}}}{\sigma_{\mathrm{initial}}}

        If the target concentration has dimensions of inverse cubic length
        (:math:`[L^{-3}]`), the value is interpreted as a number density at the
        surface for that given species and the scaling factor is computed by
        dividing that number density at the surface,
        :math:`n_{\mathrm{surface, target}}`,
        by the initial number density at the surface,
        :math:`n_{\mathrm{surface, initial}}`:

        .. math::
           f = \frac{n_{\mathrm{surface, target}}}{n_{\mathrm{surface, initial}}}

        If the target concentration has dimensions of inverse cubic length
        (:math:`[ML^{-3}]`), the value is interpreted as a mass density at the
        surface for that given species and the scaling factor is computed by
        dividing that mass density at the surface,
        :math:`\sigma_{\mathrm{surface, target}}`,
        by the initial mass density at the surface,
        :math:`\sigma_{\mathrm{surface, initial}}`:

        .. math::
           f = \frac{\sigma_{\mathrm{surface, target}}}{\sigma_{\mathrm{surface, initial}}}

        If the target concentration is dimensionless, the value is
        interpreted as a mixing ratio at the surface for that given species
        and the scaling factor is computed by dividing that mixing ratio at
        the surface,
        :math:`x_{\mathrm{surface, target}}`,
        by the initial mixing ratio at the surface,
        :math:`x_{\mathrm{surface, initial}}`:

        .. math::
           f = \frac{x_{\mathrm{target}}}{x_{\mathrm{initial}}}

    Returns
    -------
    dict
        Mapping of species (str) and scaling factors (float).
    """
    compute_initial_amount = {
        "[length]^-2": column_number_density,
        "[mass] * [length]^-2": column_mass_density,
        "[length]^-3": number_density_at_surface,
        "[mass] * [length]^-3": mass_density_at_surface,
        "": volume_mixing_ratio_at_surface,
    }
    factors = {}
    for species in concentration:
        target_amount = concentration[species]
        initial_amount = None
        for dimensions in list(compute_initial_amount.keys()):
            if target_amount.check(dimensions):
                initial_amount = compute_initial_amount[dimensions](
                    ds=ds, species=species
                )
        if initial_amount is None:
            raise ValueError(
                f"Invalid dimension for {species} concentration: {target_amount.units}."
                f"Expected dimension: [length]^-2, [mass] * [length]^-2, "
                f"[length]^-3, [mass] * [length]^-2 or no dimension."
            )
        factors[species] = _scaling_factor(
            initial_amount=initial_amount,
            target_amount=target_amount,
        )

    return factors


def human_readable(items: list[str]) -> str:
    """
    Transforms a list into readable human text.

    Example: ``["a", "b", "c"]`` -> ``"a, b and c"``

    Parameters
    ----------
    elements : list
        List.

    Returns
    -------
    str
        Human readable text.
    """
    x = f"{items[0]}"
    for s in items[1:-1]:
        x += f", {s}"
    if len(items) > 1:
        x += f" and {items[-1]}"
    return x


def rescale_concentration(
    ds: xr.Dataset, factors: dict[str, float], inplace: bool = False
) -> xr.Dataset | None:
    """
    Rescale mixing ratios in an atmosphere thermophysical properties data
    set by given factors for each species.

    Parameters
    ----------
    ds : Dataset
        Initial atmosphere thermophysical properties data set.

    factors : dict
        Mapping of species (str) and scaling factors (float).

    inplace : bool
        If ``True``, the atmosphere thermophysical properties data set object
        is modified.
        Else, a new atmosphere thermophysical properties data set object is
        returned.

    Returns
    -------
    Dataset
        Rescaled atmosphere thermophysical properties data set.
    """
    if not inplace:
        ds = ds.copy(deep=True)

    mr = ds.mr.values
    for i, species in enumerate(ds.species.values):
        if species in factors:
            mr[i] *= factors[species]

    machine_epsilon = np.finfo(float).eps
    if any(mr.sum(axis=0) > 1.0 + machine_epsilon):
        raise ValueError(
            f"Cannot rescale concentration with these factors "
            f"({factors}) because the sum of mixing ratios would "
            f"be larger than 1: {mr.sum(axis=0)}"
        )

    species = list(factors.keys())
    ds.attrs["history"] += (
        f"\n"
        f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} - "
        f"Concentration-rescaled for {human_readable(species)} - "
        f"eradiate.thermoprops.util.rescale_concentration"
    )

    return ds


@ureg.wraps(ret=None, args=(None, "km", None, None), strict=False)
def interpolate(
    ds: xr.Dataset,
    z_level: np.ndarray | pint.Quantity,
    method: str = "linear",
    conserve_columns: bool = False,
) -> xr.Dataset:
    """
    Interpolates an atmosphere thermophysical properties data set onto a
    new level altitude mesh.

    Parameters
    ----------
    ds : Dataset
        Initial atmosphere thermophysical properties data set.

    z_level : ndarray
        Level altitude mesh [km].

    method : str
        The method used to interpolate (same for all data variables).
        Choose from ``"linear"``, ``"nearest"``, ``"zero"``, ``"slinear"``, \
        ``"quadratic"``, and ``"cubic"``.

        Default: ``"linear"``.

    conserve_columns : bool
        If ``True``, multiply the number densities in the atmosphere
        thermophysical properties data set so that the column number densities
        in the initial and interpolated atmosphere thermophysical properties
        data set are in same.

        Default: ``False``.

    Returns
    -------
    Dataset
        Interpolated atmosphere thermophysical properties data set.

    Notes
    -----
    Returns a new atmosphere thermophysical properties data set object.
    """
    z_level = ureg.Quantity(z_level, "km")
    z_layer = (z_level[1:] + z_level[:-1]) / 2.0
    interpolated = ds.interp(
        z_layer=z_layer.m_as(ds.z_layer.units),
        method=method,
        kwargs=dict(fill_value="extrapolate"),
    )
    z_level_attrs = ds.z_level.attrs
    interpolated.update(
        dict(z_level=("z_level", z_level.m_as(ds.z_level.units), z_level_attrs))
    )

    if conserve_columns:
        initial_amounts = {s: column_number_density(ds, s) for s in ds.species.values}
        factors = compute_scaling_factors(
            ds=interpolated, concentration=initial_amounts
        )
        interpolated = rescale_concentration(
            ds=interpolated, factors=factors, inplace=True
        )

    return interpolated


@ureg.wraps(ret="Pa", args="K", strict=False)
def water_vapor_saturation_pressure(t: float) -> pint.Quantity:
    """
    Computes the water vapor saturation pressure over liquid water or ice, at
    the given temperature.

    Parameters
    ----------
    t : float
        Temperature [K].

    Returns
    -------
    quantity
        Water vapor saturation pressure.

    Notes
    -----
    Valid for pressures larger than the triple point pressure (~611 Pa).
    """
    try:
        import iapws
    except ModuleNotFoundError:
        warnings.warn(
            "To use the collision detection feature, you must install IAPWS.\n"
            "See instructions on https://iapws.readthedocs.io/en/latest/modules.html#installation."
        )
        raise

    if t >= 273.15:  # water is liquid
        p = ureg.Quantity(iapws.iapws97._PSat_T(t), "MPa")
    else:  # water is solid
        p = ureg.Quantity(iapws._iapws._Sublimation_Pressure(t), "MPa")
    return p.m_as("Pa")


@ureg.wraps(ret=ureg.dimensionless, args=("Pa", "K"), strict=False)
def equilibrium_water_vapor_fraction(p: float, t: float) -> pint.Quantity:
    """
    Computes the water vapor volume fraction at equilibrium, i.e., when the
    rate of condensation of water vapor equals the rate of evaporation of
    liquid water or ice, depending on the temperature.

    The water vapor volume fraction :math:`x_w` is computed with:

    .. math::
       x_w(p,T) = \\frac{p_w(T)}{p}

    where

    * :math:`p` is the pressure,
    * :math:`T` is the temperature and
    * :math:`p_w` is the water vapor saturation pressure at the given
      temperature.

    This water vapor volume fraction corresponds to a relative humidity of 100%.

    Parameters
    ----------
    p : float
        Pressure [Pa].

    t : float
        Temperature [K].

    Returns
    -------
    quantity
        Water vapor volume fraction.

    Raises
    ------
    ValueError
        If equilibrium cannot be reached in the pressure and temperature
        conditions.

    Notes
    -----
    For some values of the pressure and temperature, the equilibrium does not
    exist. An exception is raised in those cases.
    """
    p_water_vapor = water_vapor_saturation_pressure(t).magnitude
    if p_water_vapor <= p:
        return p_water_vapor / p
    else:
        raise ValueError(
            f"Equilibrium cannot be reached in these conditions (p = "
            f"{round(p, 2)} Pa, t = {round(t, 2)} K)"
        )


def make_profile_regular(profile: xr.Dataset, atol: float) -> xr.Dataset:
    """
    Converts the atmosphere thermophysical properties data set with an
    irregular altitude mesh to a profile defined over a regular altitude mesh.

    Parameters
    ----------
    profile : Dataset
        Original atmosphere thermophysical properties data set, defined over
        an irregular altitude mesh.

    atol : float
        Absolute tolerance used in the conversion of the irregular altitude
        mesh to a regular altitude mesh.

    Returns
    -------
    Dataset
        Converted atmosphere thermophysical properties data set, defined over
        a regular altitude mesh.
    """

    # compute the regular altitude nodes mesh
    regular_z_level = _to_regular(mesh=profile.z_level.values, atol=atol)

    # compute corresponding altitude centers mesh
    regular_z_layer = (regular_z_level[:-1] + regular_z_level[1:]) / 2.0

    # compute the atmospheric variables with the regular altitude mesh
    n_z = len(regular_z_layer)
    p = np.zeros(n_z)
    t = np.zeros(n_z)
    n = np.zeros(n_z)
    n_species = profile.species.size
    mr = np.zeros((n_species, n_z))
    layer = 0
    for i, z in enumerate(regular_z_layer):
        # when altitude is larger than current layer's upper bound, jump to
        # the next layer
        if z >= profile.z_level.values[layer + 1]:
            layer += 1

        p[i] = profile.p.values[layer]
        t[i] = profile.t.values[layer]
        n[i] = profile.n.values[layer]
        mr[:, i] = profile.mr.values[:, layer]

    species = profile["species"].values

    # copy attributes
    attrs = profile.attrs

    # update history
    new_line = (
        f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} - made "
        f"profile altitude mesh regular - "
        f"eradiate.thermoprops.util.make_profile_regular "
    )
    attrs["history"] += f"\n{new_line}"

    dataset = xr.Dataset(
        data_vars=dict(
            p=("z_layer", p),
            t=("z_layer", t),
            n=("z_layer", n),
            mr=(("species", "z_layer"), mr),
        ),
        coords={
            "z_layer": ("z_layer", regular_z_layer),
            "z_level": ("z_level", regular_z_level),
            "species": ("species", species),
        },
        attrs=attrs,
    )

    return dataset


def _to_regular(mesh: np.ndarray, atol: float) -> np.ndarray:
    """
    Converts an irregular altitude mesh into a regular altitude mesh.

    Parameters
    ----------
    mesh : ndarray
        Irregular altitude mesh with values sorted in increasing order.

    atol : float
        Absolute tolerance used in the conversion.

    Returns
    -------
    ndarray
        Regular altitude mesh.

    Warnings
    --------
    The algorithm is not optimized to find the approximating regular mesh with
    the smallest number of layers. Depending on the value of ``atol``, the
    resulting mesh size can be large.

    Notes
    -----
    The bound altitudes in the irregular altitude mesh remain the same in
        the output regular altitude mesh. Only the intermediate node altitudes
        are modified.
    """

    n, _ = _find_regular_params_gcd(mesh, atol)
    return np.linspace(start=mesh[0], stop=mesh[-1], num=n)


def _find_regular_params_gcd(
    mesh: np.ndarray, unit_number: float = 1.0
) -> tuple[int, float]:
    """
    Finds the parameters (number of cells, constant cell width) of the
    regular mesh that approximates the irregular input mesh.

    The algorithm finds the greatest common divisor (GCD) of all cells widths
    in the integer representation specified by the parameter ``unit_number``.
    This GCD is used to define the constant cells width of the approximating
    regular mesh.

    Parameters
    ----------
    mesh : ndarray
        1-D array with floating point values.
        Values must be sorted by increasing order.

    unit_number : float, default: 1.0
        Defines the unit used to convert the floating point numbers to integer.
        numbers.

    Returns
    -------
    int
        Number of points in the regular mesh.

    float
        Value of the constant cells width.

    Warnings
    --------
    There are no safeguards regarding how large the number of cells in the
    regular mesh can be. Use the parameter ``unit_number`` with caution.
    """

    # Convert float cell widths to integer cell widths
    eps = np.finfo(float).eps
    if unit_number >= eps:
        mesh = np.divide(mesh, unit_number).astype(int)
    else:
        raise ValueError(
            f"Parameter unit_number ({unit_number}) must be "
            f"larger than machine epsilon ({eps})."
        )
    widths = mesh[1:] - mesh[:-1]

    # Find the greatest common divisor (GCD) of all integer cell widths
    # The constant cell width in the regular mesh is given by that GCD.
    from math import gcd

    w = gcd(widths[0], widths[1])
    for x in widths[2:]:
        w = gcd(x, w)

    # Compute the number of points in the regular mesh
    total_width = mesh[-1] - mesh[0]
    n = total_width // w + 1

    return n, float(w) * unit_number


# def find_regular_params_tol(mesh, rtol=1e-3, n_cells_max=10000):
#     r"""
#     Finds the number of cells and constant cell width of the regular 1-D
#     mesh that approximates a 1-D irregular mesh the best.
#
#     Parameters
#     ----------
#     mesh : ndarray
#         Irregular 1-D mesh. Values must be sorted in increasing order.
#
#     rtol : float
#         Relative tolerance on the cells widths. This parameter controls the
#         accuracy of the approximation.
#         The parameters of the approximating regular mesh are computed so that
#         for each layer of the irregular mesh, the corresponding layer or layers
#         in the regular mesh has a width or have a total width that is not larger
#         than ``rtol`` times the width of the cell in the irregular mesh.
#
#         Default: 1e-3
#
#     n_cells_max : float
#         Maximum number of cells in the regular mesh. This parameter controls the
#         size of the resulting regular mesh.
#
#         Default: 10000
#
#     Returns
#     -------
#     int
#         Number of cells in the approximating regular mesh.
#     float
#         Constant cells width in the approximating regular mesh.
#     """
#
#     raise NotImplemented
# TODO: implement this function
